/*
 * Copyright 2010 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License version 3, as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *    Cody Russell  <crussell@canonical.com>
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <dbus/dbus-glib-bindings.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <libdbusmenu-glib/menuitem.h>
#include <libdbusmenu-glib/server.h>

#include "bridge.h"
#include "application-menu-registrar-client.h"

#define APP_MENU_DBUS_NAME   "org.ayatana.WindowMenu.Registrar"
#define APP_MENU_DBUS_OBJECT "/org/ayatana/WindowMenu/Registrar"
#define APP_MENU_INTERFACE   "org.ayatana.WindowMenu.Registrar"
#define APP_MENU_PATH        "/this/is/a/long/object/path"

static void app_menu_bridge_insert         (UbuntuMenuProxy   *proxy,
                                            GtkWidget         *shell,
                                            GtkWidget         *child,
                                            guint              position);
static gboolean app_menu_bridge_show_local (UbuntuMenuProxy   *proxy);
static void     dbus_owner_change          (DBusGProxy        *proxy,
                                            const gchar       *name,
                                            const gchar       *prev,
                                            const gchar       *new,
                                            AppMenuBridge     *bridge);
static void     toplevel_realized          (GtkWidget         *widget,
                                            gpointer           user_data);
static void     toplevel_notify_cb         (GtkWidget         *widget,
                                            GParamSpec        *pspec,
                                            UbuntuMenuProxy   *proxy);

typedef struct _AppWindowContext AppWindowContext;

struct _AppWindowContext
{
  GtkWidget        *window;
  DbusmenuServer   *server;
  DbusmenuMenuitem *root;
  gchar            *path;
  gboolean          registered;
};

struct _AppMenuBridgePrivate
{
  GHashTable *windows;    /* XID,          AppWindowContext *> */

  DBusGConnection  *session_bus;
  DBusGProxy       *dbusproxy;
  DBusGProxy       *appmenuproxy;
  gboolean          online;
};

typedef struct _RecurseContext
{
  AppMenuBridge    *bridge;
  AppWindowContext *context;

  gint count;
  DbusmenuMenuitem *stack[30];
} RecurseContext;

G_DEFINE_DYNAMIC_TYPE(AppMenuBridge, app_menu_bridge, UBUNTU_TYPE_MENU_PROXY)

static void
app_menu_bridge_setup_proxy (AppMenuBridge *bridge)
{
  bridge->priv->appmenuproxy = dbus_g_proxy_new_for_name_owner (bridge->priv->session_bus,
                                                                APP_MENU_DBUS_NAME,
                                                                APP_MENU_DBUS_OBJECT,
                                                                APP_MENU_INTERFACE,
                                                                NULL);

  bridge->priv->online = TRUE;
}

static void
app_menu_bridge_init (AppMenuBridge *bridge)
{
  bridge->priv = G_TYPE_INSTANCE_GET_PRIVATE (bridge, APP_MENU_TYPE_BRIDGE, AppMenuBridgePrivate);

  bridge->priv->windows = g_hash_table_new (g_int64_hash, g_int64_equal);

  bridge->priv->session_bus = dbus_g_bus_get(DBUS_BUS_SESSION, NULL);

  bridge->priv->dbusproxy = dbus_g_proxy_new_for_name_owner (bridge->priv->session_bus,
                                                             DBUS_SERVICE_DBUS,
                                                             DBUS_PATH_DBUS,
                                                             DBUS_INTERFACE_DBUS,
                                                             NULL);

  if (bridge->priv->dbusproxy)
    {
      dbus_g_proxy_add_signal (bridge->priv->dbusproxy,
                               "NameOwnerChanged",
                               G_TYPE_STRING,
                               G_TYPE_STRING,
                               G_TYPE_STRING,
                               G_TYPE_INVALID);

      dbus_g_proxy_connect_signal (bridge->priv->dbusproxy,
                                   "NameOwnerChanged",
                                   G_CALLBACK (dbus_owner_change),
                                   bridge,
                                   NULL);
    }

  app_menu_bridge_setup_proxy (bridge);
}

static void
app_menu_bridge_dispose (GObject *object)
{
  AppMenuBridge *bridge = APP_MENU_BRIDGE (object);

  if (bridge->priv->appmenuproxy)
    {
      g_object_unref (bridge->priv->appmenuproxy);
    }

  if (bridge->priv->dbusproxy)
    {
      g_object_unref (bridge->priv->dbusproxy);
    }

  if (bridge->priv->session_bus)
    {
      g_object_unref (bridge->priv->session_bus);
    }
}

static void
destroy_window_context (gpointer key,
                        gpointer value,
                        gpointer user_data)
{
  AppWindowContext *context = (AppWindowContext *)value;

  if (context->root)
    {
      g_object_unref (context->root);
      context->root = NULL;
    }

  if (context->server)
    {
      g_object_unref (context->server);
      context->server = NULL;
    }

  if (context->path)
    {
      g_free (context->path);
    }
}

static void
app_menu_bridge_finalize (GObject *object)
{
  AppMenuBridge *bridge = APP_MENU_BRIDGE (object);

  if (bridge->priv->windows)
    {
      g_hash_table_foreach (bridge->priv->windows,
                            destroy_window_context,
                            NULL);
      g_hash_table_destroy (bridge->priv->windows);
    }

  G_OBJECT_CLASS (app_menu_bridge_parent_class)->finalize (object);
}

static void
app_menu_bridge_class_init (AppMenuBridgeClass *class)
{
  UbuntuMenuProxyClass *proxy_class;
  GObjectClass *object_class;

  app_menu_bridge_parent_class = g_type_class_peek_parent (class);

  proxy_class = UBUNTU_MENU_PROXY_CLASS (class);
  object_class = G_OBJECT_CLASS (class);

  proxy_class->insert = app_menu_bridge_insert;
  proxy_class->show_local = app_menu_bridge_show_local;

  object_class->dispose  = app_menu_bridge_dispose;
  object_class->finalize = app_menu_bridge_finalize;

  g_type_class_add_private (class, sizeof (AppMenuBridgePrivate));
}

static void
app_menu_bridge_class_finalize (AppMenuBridgeClass *class)
{
}

static void
register_application_window (gpointer key,
                             gpointer value,
                             gpointer user_data)
{
  AppWindowContext *context     = (AppWindowContext *)value;
  GtkWidget        *widget      = context->window;
  AppMenuBridge    *bridge     = APP_MENU_BRIDGE (user_data);

  if (!context->registered && context->server != NULL && context->root != NULL && GTK_IS_WINDOW (widget) && bridge->priv->appmenuproxy != NULL)
    {
      dbusmenu_server_set_root (context->server, context->root);
      org_ayatana_WindowMenu_Registrar_register_window (bridge->priv->appmenuproxy,
                                                        GDK_WINDOW_XID (gtk_widget_get_window (widget)),
                                                        context->path,
                                                        NULL);

      context->registered = TRUE;
    }
}

static void
register_application_windows (AppMenuBridge *bridge)
{
  g_hash_table_foreach (bridge->priv->windows,
                        register_application_window,
                        bridge);
}

static void
unregister_application_window (gpointer key,
                               gpointer value,
                               gpointer user_data)
{
  AppWindowContext *context = (AppWindowContext *)value;

  context->registered = FALSE;
}

static void
unregister_application_windows (AppMenuBridge *bridge)
{
  g_hash_table_foreach (bridge->priv->windows,
                        unregister_application_window,
                        bridge);
}

static void
app_menu_bridge_proxy_vanished (AppMenuBridge *bridge)
{
  if (bridge->priv->online)
    g_print ("VANISH\n");

  bridge->priv->online = FALSE;

  unregister_application_windows (bridge);
}

static void
app_menu_bridge_proxy_appeared (AppMenuBridge *bridge)
{
  g_print ("APPEAR\n");

  bridge->priv->online = TRUE;

  app_menu_bridge_setup_proxy (bridge);

  register_application_windows (bridge);
}

static void
dbus_owner_change (DBusGProxy    *proxy,
                   const gchar   *name,
                   const gchar   *prev,
                   const gchar   *new,
                   AppMenuBridge *bridge)
{
  if (strlen (new) > 0 && strlen (prev) == 0)
    {
      if (g_strcmp0 (name, APP_MENU_DBUS_NAME) == 0)
        {
          app_menu_bridge_proxy_appeared (bridge);
        }
    }
  else if (strlen (new) == 0 && strlen (prev) > 0)
    {
      app_menu_bridge_proxy_vanished (bridge);
    }
}

static GtkWidget *
find_menu_label (GtkWidget *widget)
{
  GtkWidget *label = NULL;

  if (GTK_IS_LABEL (widget))
    return widget;

  if (GTK_IS_CONTAINER (widget))
    {
      GList *children;
      GList *l;

      children = gtk_container_get_children (GTK_CONTAINER (widget));

      for (l = children; l; l = l->next)
        {
          label = find_menu_label (l->data);

          if (label)
            break;
        }

      g_list_free (children);
    }

  return label;
}

static const gchar *
get_menu_label_text (GtkWidget *menuitem)
{
  GtkWidget *label = find_menu_label (menuitem);

  if (label)
    return gtk_label_get_text (GTK_LABEL (label));

  return NULL;
}

static void
item_activated (DbusmenuMenuitem *item, guint timestamp, gpointer user_data)
{
  GtkWidget *child;

  if (user_data != NULL)
    {
      child = (GtkWidget *)user_data;

      if (GTK_IS_MENU_ITEM (child))
        {
          gtk_menu_item_activate (GTK_MENU_ITEM (child));
        }
    }
}

static void
widget_notify_cb (GtkWidget  *widget,
                  GParamSpec *pspec,
                  gpointer    data)
{
  DbusmenuMenuitem *child = (DbusmenuMenuitem *)data;

  if (pspec->name == g_intern_static_string ("sensitive"))
    {
      dbusmenu_menuitem_property_set_bool (child,
                                           DBUSMENU_MENUITEM_PROP_ENABLED,
                                           gtk_widget_get_sensitive (widget));
    }
  else if (pspec->name == g_intern_static_string ("label"))
    {
      dbusmenu_menuitem_property_set (child,
                                      DBUSMENU_MENUITEM_PROP_LABEL,
                                      gtk_menu_item_get_label (GTK_MENU_ITEM (widget)));
    }
  else if (pspec->name == g_intern_static_string ("visible"))
    {
      dbusmenu_menuitem_property_set_bool (child,
                                           DBUSMENU_MENUITEM_PROP_VISIBLE,
                                           gtk_widget_get_visible (widget));
    }
}

static void
checkbox_toggled (GtkWidget *widget, DbusmenuMenuitem *mi)
{
  dbusmenu_menuitem_property_set_int (mi,
                                      DBUSMENU_MENUITEM_PROP_TOGGLE_STATE,
                                      gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (widget)) ? DBUSMENU_MENUITEM_TOGGLE_STATE_CHECKED : DBUSMENU_MENUITEM_TOGGLE_STATE_UNCHECKED);
}

static DbusmenuMenuitem *
construct_dbusmenu_for_widget (GtkWidget *widget)
{
  DbusmenuMenuitem *mi = dbusmenu_menuitem_new ();

  if (GTK_IS_MENU_ITEM (widget))
    {
      if (GTK_IS_SEPARATOR_MENU_ITEM (widget))
        {
          dbusmenu_menuitem_property_set (mi,
                                          "type",
                                          "separator");
        }
      else
        {
          if (GTK_IS_CHECK_MENU_ITEM (widget))
            {
              dbusmenu_menuitem_property_set (mi,
                                              DBUSMENU_MENUITEM_PROP_TOGGLE_TYPE,
                                              GTK_IS_RADIO_MENU_ITEM (widget) ? DBUSMENU_MENUITEM_TOGGLE_RADIO : DBUSMENU_MENUITEM_TOGGLE_CHECK);

              dbusmenu_menuitem_property_set_int (mi,
                                                  DBUSMENU_MENUITEM_PROP_TOGGLE_STATE,
                                                  gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (widget)) ? DBUSMENU_MENUITEM_TOGGLE_STATE_CHECKED : DBUSMENU_MENUITEM_TOGGLE_STATE_UNCHECKED);

              g_signal_connect (widget,
                                "toggle",
                                G_CALLBACK (checkbox_toggled),
                                mi);
            }

          dbusmenu_menuitem_property_set (mi,
                                          "label",
                                          get_menu_label_text (widget));

          dbusmenu_menuitem_property_set_bool (mi,
                                               DBUSMENU_MENUITEM_PROP_VISIBLE,
                                               gtk_widget_get_visible (widget));

          dbusmenu_menuitem_property_set_bool (mi,
                                               DBUSMENU_MENUITEM_PROP_ENABLED,
                                               gtk_widget_get_sensitive (widget));

          g_signal_connect (G_OBJECT (widget),
                            "notify",
                            G_CALLBACK (widget_notify_cb),
                            mi);

          g_signal_connect (G_OBJECT (mi),
                            "item_activated",
                            G_CALLBACK (item_activated),
                            widget);
        }
    }

  return mi;
}

static void
rebuild_item (GtkWidget      *widget,
              RecurseContext *recurse)
{
  if (GTK_IS_CONTAINER (widget))
    {
      gboolean increment = GTK_IS_MENU_BAR (widget) || GTK_IS_MENU_ITEM (widget);

      if (increment)
        recurse->count++;

      if (recurse->count > -1 && increment)
        {
          recurse->stack[recurse->count] = construct_dbusmenu_for_widget (widget);

          if (recurse->count > 0)
            {
              if (recurse->count == 1)
                dbusmenu_menuitem_child_append (recurse->stack[recurse->count - 1],
                                                recurse->stack[recurse->count]);
              else
                dbusmenu_menuitem_child_prepend (recurse->stack[recurse->count - 1],
                                                 recurse->stack[recurse->count]);
            }
        }

      gtk_container_foreach (GTK_CONTAINER (widget),
                             (GtkCallback)rebuild_item,
                             recurse);

      if (GTK_IS_MENU_ITEM (widget))
        {
          GtkWidget *menu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (widget));

          if (menu != NULL)
            {
              rebuild_item (menu, recurse);
            }
        }

      if (increment)
        recurse->count--;
    }
}

static void
rebuild_window_items (AppMenuBridge *bridge,
                      GtkWidget     *toplevel)
{
  XID xid;
  AppWindowContext *context;
  RecurseContext recurse;

  if (!GTK_IS_WINDOW (toplevel))
    {
      g_signal_connect (G_OBJECT (toplevel),
                        "notify",
                        G_CALLBACK (toplevel_notify_cb),
                        bridge);

      return;
    }
  else
    {
      gboolean no_proxy;

      g_object_get (G_OBJECT (toplevel),
                    "ubuntu-no-proxy", &no_proxy,
                    NULL);

      if (no_proxy)
        {
          return;
        }
    }

  if (!gtk_widget_get_realized (toplevel))
    {
      g_signal_connect (toplevel, "realize",
                        G_CALLBACK (toplevel_realized),
                        bridge);

      return;
    }

  xid = GDK_WINDOW_XID (gtk_widget_get_window (toplevel));
  context = g_hash_table_lookup (bridge->priv->windows, &xid);

  if (!context)
    {
      context = g_new0 (AppWindowContext, 1);
      g_hash_table_insert (bridge->priv->windows, &xid, context);
    }

  context->window = toplevel;

  if (!context->path)
    context->path = g_strdup_printf ("/org/ayatana/menu/%X", (guint)xid);

  if (!context->server)
    context->server = dbusmenu_server_new (context->path);

  recurse.bridge  = bridge;
  recurse.context = context;
  recurse.count   = -1;

  gtk_container_foreach (GTK_CONTAINER (toplevel),
                         (GtkCallback)rebuild_item,
                         &recurse);

  context->root = recurse.stack[0];

  dbusmenu_server_set_root (context->server, context->root);
}


static void
toplevel_realized (GtkWidget *widget,
                   gpointer   user_data)
{
  AppMenuBridge *bridge = APP_MENU_BRIDGE (user_data);

  if (GTK_IS_WINDOW (widget))
    {
      rebuild_window_items (bridge, widget);
      register_application_windows (bridge);

      return;
    }
}

static void
toplevel_notify_cb (GtkWidget       *widget,
                    GParamSpec      *pspec,
                    UbuntuMenuProxy *proxy)
{
  if (pspec->name == g_intern_static_string ("parent"))
    {
      GtkWidget *toplevel = gtk_widget_get_toplevel (widget);
      AppMenuBridge *bridge = APP_MENU_BRIDGE (proxy);

      rebuild_window_items (bridge, toplevel);
    }
}

static void
attach_notify_cb (GtkWidget     *widget,
                  GParamSpec    *pspec,
                  AppMenuBridge *bridge)
{
  if (pspec->name == g_intern_static_string ("attach-widget"))
    {
      GtkWidget *attach = NULL;

      g_object_get (widget, "attach-widget", &attach, NULL);

      rebuild_window_items (bridge, attach);
    }
}

static void
app_menu_bridge_insert (UbuntuMenuProxy *proxy,
                        GtkWidget       *parent,
                        GtkWidget       *child,
                        guint            position)
{
  AppMenuBridge        *bridge;
  AppMenuBridgePrivate *priv;
  GtkWidget            *toplevel = NULL;
  gboolean              append = FALSE;

  if (GTK_IS_TEAROFF_MENU_ITEM (child))
    return;

  bridge = APP_MENU_BRIDGE (proxy);
  priv = bridge->priv;

  /*
  g_print ("INSERT (parent %s %p, child %s %p\n",
           G_OBJECT_TYPE_NAME (parent), parent,
           G_OBJECT_TYPE_NAME (child), child);
  */

  toplevel = gtk_widget_get_toplevel (parent);

  if (GTK_IS_MENU_BAR (parent))
    {
      g_signal_connect (G_OBJECT (toplevel),
                        "notify",
                        G_CALLBACK (toplevel_notify_cb),
                        proxy);

      append = TRUE;
    }
  else if (GTK_IS_MENU (parent))
    {
      GtkWidget *attach;

      g_object_get (parent, "attach-widget", &attach, NULL);

      if (attach == NULL)
        {
          g_signal_connect (G_OBJECT (parent),
                            "notify",
                            G_CALLBACK (attach_notify_cb),
                            bridge);
          return;
        }
    }

  if (GTK_IS_WINDOW (toplevel))
    {
      g_signal_connect (toplevel, "realize",
                        G_CALLBACK (toplevel_realized),
                        bridge);
    }
}

static gboolean
app_menu_bridge_show_local (UbuntuMenuProxy *proxy)
{
  const gchar *env = g_getenv ("APPMENU_DISPLAY_BOTH");

  return (g_strcmp0 (env, "1") == 0);
}

/* crude blacklist to avoid to much patching for now */
static gboolean
app_menu_brige_shouldnt_load (void)
{
	const char *prg = g_get_prgname ();

	if (g_strrstr (prg, "indicator-applet") != NULL) {
		g_print ("*** Warning: menu proxy blacklisting: %s\n", prg);
		return TRUE;
	}

	return FALSE;
}


G_MODULE_EXPORT void
menu_proxy_module_load (UbuntuMenuProxyModule *module)
{
  static gboolean registered = FALSE;

  /* Prevent well-known applications to re-export
	 their own dbusmenus */
  if (app_menu_brige_shouldnt_load ())
	  return;

  if (!registered)
    {
      app_menu_bridge_register_type (G_TYPE_MODULE (module));

      registered = TRUE;
    }
}

G_MODULE_EXPORT void
menu_proxy_module_unload (UbuntuMenuProxyModule *module)
{
}
